import sys
import os
import datetime
import time
import zipfile
import tempfile
try:
    from distutils import dir_util
except:
    pass
import shutil
import json
import traceback

from werkzeug._compat import to_unicode
from ..common import consts, file_read, file_write, error_message
from ..db.db_modules import SQLITE, FIREBIRD, get_db_module
from .builder import items_insert_sql, items_update_sql, items_delete_sql
from .builder import indices_insert_sql, indices_delete_sql

from .export_metadata import metadata_items

def import_metadata(admin, file_name, from_client=False):
    imp = MetaDataImport(admin, file_name, from_client)
    return imp.import_metadata()

class MetaDataImport(object):
    def __init__(self, task, file_name, from_client):
        self.task = task
        self.file_name = os.path.join(task.work_dir, os.path.normpath(file_name))
        self.from_client = from_client
        self.client_log = ''
        self.server_log = ''
        self.success = True
        self.error = None
        self.tmpdir = None
        self.new_items = {}
        self.old_items = {}
        self.db_type = task.task_db_type
        self.new_db_type = None
        self.db_sql = None
        self.adm_sql = None
        self.db_module = task.task_db_module
        self.items_hidden_fields = ['f_gen_name']
        self.params_hidden_fields = [
                'f_safe_mode', 'f_debugging', 'f_modification',
                'f_client_modified', 'f_server_modified',
                'f_build_version', 'f_params_version',
                'f_maintenance', 'f_import_delay', 'f_production'
            ]

    def import_metadata(self):
        self.check_can_import()
        self.prepare_data()
        self.check_data_integrity()
        self.analize_data()
        self.wait_ready()
        self.import_databases()
        self.copy_files()
        self.update_logs()
        self.tidy_up()
        return self.success, self.error, self.client_log

    def check_can_import(self):
        if self.db_type == SQLITE and not self.project_empty():
            self.success = False
            self.error = 'Metadata can not be imported into an existing SQLITE project'
            self.show_error(self.error)

    def update_indexes(self):
        if self.new_db_type == FIREBIRD or self.db_type == FIREBIRD:
            item = self.new_items['sys_indices']
            for it in item:
                if it.f_fields_list.value:
                    field_list = it.load_index_fields(it.f_fields_list.value)
                    desc = it.descending.value
                    if field_list:
                        it.edit()
                        if self.new_db_type == FIREBIRD:
                            l = []
                            for f in field_list:
                                l.append([f[0], desc])
                            field_list = l
                        elif self.db_type == FIREBIRD:
                            desc = field_list[0][1]
                        it.descending.value = desc
                        it.f_fields_list.value = it.store_index_fields(field_list)
                        it.post()

    def update_item_idents(self, item_name, field_names, case):
        item = self.new_items[item_name]
        fields = []
        for field_name in field_names:
            fields.append(item.field_by_name(field_name))
        item.log_changes = False
        for it in item:
            it.edit()
            for field in fields:
                field.value = case(field.value)
            it.post()

    def update_idents(self):
        case = get_db_module(self.db_type).identifier_case
        self.update_item_idents('sys_items', ['f_table_name', 'f_gen_name'], case)
        self.update_item_idents('sys_fields', ['f_db_field_name'], case)
        self.update_item_idents('sys_indices', ['f_index_name'], case)
        self.update_indexes()

    def prepare_data(self):
        if self.success:
            self.show_progress(self.task.language('import_reading_data'))
            try:
                self.tmpdir =  tempfile.mkdtemp()
                with zipfile.ZipFile(self.file_name) as z:
                    z.extractall(self.tmpdir)
                    file_name = os.path.join(self.tmpdir, 'task.dat')
                    data = file_read(file_name)
                    data_lists = json.loads(data)
                    for item_name in metadata_items:
                        item = self.task.item_by_name(item_name)
                        self.task.execute('DELETE FROM "%s" WHERE "DELETED" = 1' % item.table_name)
                        old_item = item.copy(handlers=False)
                        old_item.soft_delete = False
                        old_item.open(expanded=False)
                        field_names, dataset = self.get_dataset(old_item, data_lists)
                        new_item = item.copy(handlers=False)
                        new_item.open(expanded=False, fields=field_names, open_empty=True)
                        new_item._dataset = dataset
                        self.new_items[item.item_name] = new_item
                        self.old_items[item.item_name] = old_item
                    os.remove(file_name)
                    self.new_db_type = data_lists.get('db_type')
                    if self.new_db_type != self.db_type:
                        self.update_idents()
            except Exception as e:
                self.task.log.exception(e)
                self.success = False
                self.show_error(e)

    def get_dataset(self, item, data_lists):
        ns = []
        ds = []
        dl = data_lists.get(item.item_name)
        if dl:
            field_names = data_lists[item.item_name]['fields']
            dataset = data_lists[item.item_name]['records']
            for d in dataset:
                ds.append([])
            for i, f in enumerate(field_names):
                if item.field_by_name(f):
                    ns.append(f)
                    for j, d in enumerate(dataset):
                        ds[j].append(dataset[j][i])
        else:
            for f in item.fields:
                ns.append(f.field_name)
        return ns, ds

    def check_data_integrity(self):
        if self.success:
            self.show_progress(self.task.language('import_checking_integrity'))
            errors = []
            new = self.new_items['sys_items']
            old = self.old_items['sys_items']
            compare = self.compare_items(old, new)
            for it in old:
                o, n = compare[old.id.value]
                if o and n:
                    new.locate('id', old.id.value)
                    if old.type_id.value != new.type_id.value:
                        errors.append('Items with ID %s (%s, %s) have different type values' % \
                        (old.id.value, old.f_item_name.value, new.f_item_name.value))
                    elif old.f_table_name.value and old.f_table_name.value.upper() != new.f_table_name.value.upper():
                        errors.append('Items with ID %s (%s, %s) have different database tables (%s, %s)' % \
                        (old.id.value, old.f_item_name.value, new.f_item_name.value, old.f_table_name.value, new.f_table_name.value))
            if len(errors):
                self.error = "\n".join(errors)
                self.success = False
                self.show_error(self.error)

    def compare_items(self, old, new, owner_id=None):
        result = {}
        for it in old:
            result[it.id.value] = [True, False]
        for it in new:
            if not owner_id or owner_id == it.owner_rec_id.value:
                info = result.get(it.id.value)
                if info:
                    info[1] = True
                else:
                    result[it.id.value] = [False, True]
        return result

    def analize_data(self):
        if self.success:
            self.show_progress(self.task.language('import_analyzing'))
            try:
                task = self.task
                db_sql = []
                adm_sql = []
                deltas = {}

                delta = self.get_delta('sys_indices', options=['delete'])
                for d in delta:
                    table_name = self.get_table_name(d.owner_rec_id.value)
                    if table_name:
                        db_sql.append(indices_delete_sql(task.sys_indices, d))
                adm_sql.append(delta.apply_sql())

                delta = self.get_delta('sys_items', 'sys_fields')
                self.check_generator(task.sys_items, delta)
                for d in delta:
                    if d.rec_inserted():
                        db_sql.append(items_insert_sql(task.sys_items, d,
                            new_fields=self.get_new_fields(d.id.value)))
                    elif d.rec_modified():
                        db_sql.append(items_update_sql(task.sys_items, d))
                    elif d.rec_deleted():
                        db_sql.append(items_delete_sql(task.sys_items, d))

                self.refresh_old_item('sys_items')
                delta = self.get_delta('sys_items')
                self.check_generator(task.sys_items, delta)
                adm_sql.append(delta.apply_sql())

                self.refresh_old_item('sys_fields')
                delta = self.get_delta('sys_fields')
                adm_sql.append(delta.apply_sql())

                self.refresh_old_item('sys_indices')
                delta = self.get_delta('sys_indices', options=['update', 'insert'])
                for d in delta:
                    table_name = self.get_table_name(d.owner_rec_id.value)
                    if table_name:
                        if d.rec_inserted():
                            db_sql.append(indices_insert_sql(
                                task.sys_indices,
                                d, table_name,
                                self.get_new_fields(d.owner_rec_id.value),
                                foreign_key_dict=self.get_foreign_key_dict(d)
                                )
                            )
                        elif d.rec_deleted():
                            db_sql.append(indices_delete_sql(task.sys_indices, d))
                adm_sql.append(delta.apply_sql())

                for item_name in ['sys_filters', 'sys_report_params', 'sys_roles', 'sys_params',
                    'sys_privileges', 'sys_lookup_lists']:
                    delta = self.get_delta(item_name)
                    adm_sql.append(delta.apply_sql())

                self.db_sql = self.sqls_to_list(db_sql)
                self.adm_sql = self.sqls_to_list(adm_sql)
            except Exception as e:
                self.task.log.exception(e)
                self.success = False
                self.show_error(e)

    def get_table_name(self, item_id):
        items = self.new_items['sys_items']
        if items.locate('id', item_id):
            if not items.f_virtual_table.value:
                return items.f_table_name.value

    def get_foreign_key_dict(self, ind):
        dic = None
        if ind.f_foreign_index.value:
            dic = {}
            fields = self.new_items['sys_fields']
            fields.locate('id', ind.f_foreign_field.value)
            dic['key'] = fields.f_db_field_name.value
            ref_id = fields.f_object.value
            items = self.new_items['sys_items']
            items.locate('id', ref_id)
            dic['ref'] = items.f_table_name.value
            primary_key = items.f_primary_key.value
            fields.locate('id', primary_key)
            dic['primary_key'] = fields.f_db_field_name.value
        return dic

    def get_new_fields(self, item_id):
        result = []
        items = self.new_items['sys_items']
        if items.locate('id', item_id):
            parent_id = items.parent.value
        new_fields = self.new_items['sys_fields']
        for field in new_fields:
            if field.owner_rec_id.value in [item_id, parent_id]:
                if not field.f_master_field.value:
                    dic = {}
                    dic['id'] = field.id.value
                    dic['field_name'] = field.f_db_field_name.value
                    dic['data_type'] = field.f_data_type.value
                    dic['size'] = field.f_size.value
                    dic['default_value'] = ''#field.f_default_value.value
                    dic['primary_key'] = field.id.value == items.f_primary_key.value
                    result.append(dic)
        return result

    def can_copy_field(self, field):
        if field.owner.item_name == 'sys_params':
            if field.field_name in self.params_hidden_fields:
                return False
        if field.owner.item_name == 'sys_items':
            if field.field_name in self.items_hidden_fields:
                return False
        return True

    def copy_record(self, old, new):
        for old_field in old.fields:
            if self.can_copy_field(old_field):
                new_field = new.field_by_name(old_field.field_name)
                if new_field:
                    old_field.value = new_field.raw_value

    def update_item(self, item_name, detail_name=None,
        options=['update', 'insert', 'delete'], owner=None):

        new = self.new_items[item_name]
        if owner:
            old = owner.detail_by_name(item_name)
            old.open(expanded=False)
        else:
            old = self.old_items[item_name]
        owner_id = None
        if owner:
            owner_id = owner.id.value

        compare = self.compare_items(old, new, owner_id)

        if 'delete' in options:
            old.first()
            while not old.eof():
                if not owner_id or owner_id == old.owner_rec_id.value:
                    o, n = compare[old.id.value]
                    if o and not n:
                        old.delete()
                    else:
                        old.next()
                else:
                    old.next()

        if 'update' in options:
            new_ids = {}
            for it in new:
                new_ids[new.id.value] = new.rec_no
            for it in old:
                if not owner_id or owner_id == it.owner_rec_id.value:
                    o, n = compare[old.id.value]
                    if o and n:
                        rec = new_ids.get(old.id.value)
                        if rec is not None:
                            new.rec_no = rec
                            old.edit()
                            self.copy_record(old, new)
                            if detail_name:
                                self.update_item(detail_name, owner=old)
                            old.post()

        if 'insert' in options:
            for it in new:
                if not owner_id or owner_id == it.owner_rec_id.value:
                    o, n = compare[new.id.value]
                    if not o and n:
                        old.append()
                        self.copy_record(old, new)
                        if detail_name:
                            self.update_item(detail_name, owner=old)
                        old.post()

        return old

    def get_delta(self, item_name, detail_name=None, options=['update', 'insert', 'delete']):
        item = self.update_item(item_name, detail_name, options)
        return item.delta()

    def check_generator(self, item, delta):
        for d in delta:
            module = get_db_module(self.db_type)
            if d.rec_inserted() and module.NEED_GENERATOR and \
                d.f_primary_key.value and not d.f_gen_name.value and \
                not d.f_virtual_table.value:
                case = module.identifier_case
                d.edit()
                d.f_gen_name.value = case('%s_SEQ' % d.f_table_name.value)
                d.post()

    def refresh_old_item(self, item_name):
        item = self.task.item_by_name(item_name).copy(handlers=False)
        item.open(expanded=False)
        self.old_items[item_name] = item

    def wait_ready(self):
        if self.success:
            if self.from_client:
                self.show_progress(self.task.language('import_waiting_close'))
                request_count = int(self.from_client)
                if consts.IMPORT_DELAY:
                    time.sleep(consts.IMPORT_DELAY)
                else:
                    while True:
                        i = 0
                        if self.task.app._busy > request_count:
                            time.sleep(0.1)
                            i += 1
                            if i > 3000:
                                break
                        else:
                            break

    def import_databases(self):
        if self.success:
            self.show_progress(self.task.language('import_changing_db'))
            connection = self.execute_ddl()
            try:
                if self.success:
                    admin_name = os.path.join(self.task.work_dir, 'admin.sqlite')
                    tmp_admin_name = os.path.join(self.task.work_dir, '_admin.sqlite')
                    if self.db_module.DDL_ROLLBACK:
                        shutil.copyfile(admin_name, tmp_admin_name)
                    self.show_progress(self.task.language('import_changing_admin'))
                    result, error = self.task.execute(self.adm_sql)
                    self.error = error
                    if self.error:
                        self.success = False
                    if self.db_module.DDL_ROLLBACK:
                        if self.success:
                            connection.commit()
                            os.remove(tmp_admin_name)
                        else:
                            os.rename(tmp_admin_name, admin_name)
                            connection.rollback()
            finally:
                connection.close();
            consts.read_settings()
            consts.read_language()

    def execute_ddl(self):
        task = self.task
        info = []
        error = None
        connection = None
        try:
            connection = self.db_module.connect(
                task.task_db_database,
                task.task_db_user,
                task.task_db_password,
                task.task_db_host,
                task.task_db_port,
                task.task_db_encoding,
                task.task_db_server)
            if self.db_sql:
                cursor = connection.cursor()
                for sql in self.db_sql:
                    try:
                        cursor.execute(sql)
                    except Exception as x:
                        self.task.log.exception('Error: %s query: %s' % (x, sql))
                        error = error_message(x)
                    info.append({'sql': sql, 'error': error})
                    if error and self.db_module.DDL_ROLLBACK:
                        break
                if self.db_module.DDL_ROLLBACK:
                    if error:
                        self.success = False
                else:
                    connection.commit()
        except Exception as x:
            error = str(x)
            info.append({'error': error})
        self.show_info(info)
        return connection

    def sqls_to_list(self, sqls, result=None):
        if result is None:
            result = []
        for sql in sqls:
            if sql:
                if type(sql) == list:
                    self.sqls_to_list(sql, result)
                else:
                    result.append(sql)
        return result

    def update_logs(self):
        if self.success:
            result = self.task.language('import_success')
            if self.error:
                result = self.task.language('import_errors')
        else:
            result = self.task.language('import_failed')
        self.task.log.info(result)

        self.server_log = '%s\n\n%s' % (result.upper(), self.server_log)
        log_dir = os.path.join(self.task.work_dir, 'logs')
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
        log_file_name = os.path.join(log_dir, 'import_%s.log' % datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S"))
        file_write(log_file_name, self.server_log)

        if self.success:
            message = '<h3 class="text-center">%s</h3>' % result
        else:
            message = '<h3 class="text-center text-error">%s</h3>' % result
        self.client_log = '%s<h4 class="text-info">%s</h4><div>%s</div>' % \
            (message, self.task.language('import_log'), self.client_log)


    def copy_files(self):
        if self.success:
            self.show_progress(self.task.language('import_copying'))
            try:
                shutil.copytree(self.tmpdir, self.task.work_dir,
                    copy_function=shutil.copyfile, dirs_exist_ok=True)
            except:
                dir_util.copy_tree(self.tmpdir, self.task.work_dir)

    def tidy_up(self):
        self.show_progress(self.task.language('import_deleteing_files'))
        try:
            if self.tmpdir and os.path.exists(self.tmpdir):
                shutil.rmtree(self.tmpdir)
            if self.success or self.from_client:
                os.remove(self.file_name)
        except Exception as e:
            self.task.log.exception(e)
            self.show_error(e)


    def show_progress(self, string):
        self.task.log.info(string)
        self.client_log += '<h5>' + string + '</h5>'
        self.server_log += '\n%s\n' % string

    def show_info(self, errors):
        for info in errors:
            sql = info.get('sql')
            error = info.get('error')
            if sql:
                self.task.log.info(sql)
                if error:
                    self.client_log += '<div class="text-error" style="margin-bottom: 10px; margin-left: 20px;">' + sql + '</div>'
                else:
                    self.client_log += '<div style="margin-bottom: 10px; margin-left: 20px;">' + sql + '</div>'
                self.server_log += '\n%s' % sql
            if error:
                self.show_error(error)

    def show_error(self, error):
        mess = self.format_error(error_message(error))
        self.client_log += '<div class="text-error" style="margin-left: 40px;">' + mess + '</div>'
        self.server_log += '\n%s' % error

    def format_error(self, error):
        try:
            arr = str(error).split('\n')
            lines = []
            for line in arr:
                line = line.replace('\t', ' ')
                spaces = 0
                for ch in line:
                    if ch == ' ':
                        spaces += 1
                    else:
                        break
                if spaces:
                    line = '<span style="white-space: pre; margin-left: %spx">%s</span>' % (10 * (spaces - 1), line)
                else:
                    line = '<span style="white-space: pre;">%s</span>' % line
                lines.append(line)
            result = '<br>'.join(lines)
            return '<div class="text-error">%s</div>' % result
        except:
            return error

    def project_empty(self):
        items = self.task.sys_items.copy(handlers=False)
        items.open(fields=['id', 'f_table_name'])
        for i in items:
            if i.f_table_name.value:
                return False
        return True



# sql/roles.py
# Copyright (C) 2005-2020 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php


class SQLRole(object):
    """Define a "role" within a SQL statement structure.

    Classes within SQL Core participate within SQLRole hierarchies in order
    to more accurately indicate where they may be used within SQL statements
    of all types.

    .. versionadded:: 1.4

    """


class UsesInspection(object):
    pass


class ColumnArgumentRole(SQLRole):
    _role_name = "Column expression"


class ColumnArgumentOrKeyRole(ColumnArgumentRole):
    _role_name = "Column expression or string key"


class ColumnListRole(SQLRole):
    """Elements suitable for forming comma separated lists of expressions."""


class TruncatedLabelRole(SQLRole):
    _role_name = "String SQL identifier"


class ColumnsClauseRole(UsesInspection, ColumnListRole):
    _role_name = "Column expression or FROM clause"

    @property
    def _select_iterable(self):
        raise NotImplementedError()


class LimitOffsetRole(SQLRole):
    _role_name = "LIMIT / OFFSET expression"


class ByOfRole(ColumnListRole):
    _role_name = "GROUP BY / OF / etc. expression"


class OrderByRole(ByOfRole):
    _role_name = "ORDER BY expression"


class StructuralRole(SQLRole):
    pass


class StatementOptionRole(StructuralRole):
    _role_name = "statement sub-expression element"


class WhereHavingRole(StructuralRole):
    _role_name = "SQL expression for WHERE/HAVING role"


class ExpressionElementRole(SQLRole):
    _role_name = "SQL expression element"


class ConstExprRole(ExpressionElementRole):
    _role_name = "Constant True/False/None expression"


class LabeledColumnExprRole(ExpressionElementRole):
    pass


class BinaryElementRole(ExpressionElementRole):
    _role_name = "SQL expression element or literal value"


class InElementRole(SQLRole):
    _role_name = (
        "IN expression list, SELECT construct, or bound parameter object"
    )


class FromClauseRole(ColumnsClauseRole):
    _role_name = "FROM expression, such as a Table or alias() object"

    _is_subquery = False

    @property
    def _hide_froms(self):
        raise NotImplementedError()


class StrictFromClauseRole(FromClauseRole):
    # does not allow text() or select() objects
    pass


class AnonymizedFromClauseRole(StrictFromClauseRole):
    # calls .alias() as a post processor

    def _anonymous_fromclause(self, name=None, flat=False):
        """A synonym for ``.alias()`` that is only present on objects of this
        role.

        This is an implicit assurance of the target object being part of the
        role where anonymous aliasing without any warnings is allowed,
        as opposed to other kinds of SELECT objects that may or may not have
        an ``.alias()`` method.

        The method is used by the ORM but is currently semi-private to
        preserve forwards-compatibility.

        """
        return self.alias(name=name, flat=flat)


class CoerceTextStatementRole(SQLRole):
    _role_name = "Executable SQL, text() construct, or string statement"


class StatementRole(CoerceTextStatementRole):
    _role_name = "Executable SQL or text() construct"


class ReturnsRowsRole(StatementRole):
    _role_name = (
        "Row returning expression such as a SELECT, a FROM clause, or an "
        "INSERT/UPDATE/DELETE with RETURNING"
    )


class SelectStatementRole(ReturnsRowsRole):
    _role_name = "SELECT construct or equivalent text() construct"

    def subquery(self):
        raise NotImplementedError(
            "All SelectStatementRole objects should implement a "
            ".subquery() method."
        )


class HasCTERole(ReturnsRowsRole):
    pass


class CompoundElementRole(SQLRole):
    """SELECT statements inside a CompoundSelect, e.g. UNION, EXTRACT, etc."""

    _role_name = (
        "SELECT construct for inclusion in a UNION or other set construct"
    )


class DMLRole(StatementRole):
    pass


class DMLColumnRole(SQLRole):
    _role_name = "SET/VALUES column expression or string key"


class DMLSelectRole(SQLRole):
    """A SELECT statement embedded in DML, typically INSERT from SELECT """

    _role_name = "SELECT statement or equivalent textual object"


class DDLRole(StatementRole):
    pass


class DDLExpressionRole(StructuralRole):
    _role_name = "SQL expression element for DDL constraint"


class DDLConstraintColumnRole(SQLRole):
    _role_name = "String column name or column expression for DDL constraint"


class DDLReferredColumnRole(DDLConstraintColumnRole):
    _role_name = (
        "String column name or Column object for DDL foreign key constraint"
    )



# sqlalchemy/events.py
# Copyright (C) 2005-2020 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

"""Core event interfaces."""

from .engine.events import ConnectionEvents  # noqa
from .engine.events import DialectEvents  # noqa
from .pool.events import PoolEvents  # noqa
from .sql.base import SchemaEventTarget  # noqa
from .sql.events import DDLEvents  # noqa



# -*- coding: utf-8 -*-
# Copyright JS Foundation and other contributors, https://js.foundation/
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import, unicode_literals


def toDict(value):
    from .visitor import ToDictVisitor
    return ToDictVisitor().visit(value)


class Array(list):
    pass


class Object(object):
    def toDict(self):
        from .visitor import ToDictVisitor
        return ToDictVisitor().visit(self)

    def __repr__(self):
        from .visitor import ReprVisitor
        return ReprVisitor().visit(self)

    def __getattr__(self, name):
        return None



import psycopg2
from werkzeug._compat import PY2, iteritems, text_type, to_bytes, to_unicode

psycopg2.extensions.register_type(psycopg2.extensions.UNICODE)

DATABASE = 'POSTGRESQL'
NEED_DATABASE_NAME = True
NEED_LOGIN = True
NEED_PASSWORD = True
NEED_ENCODING = True
NEED_HOST = True
NEED_PORT = True
CAN_CHANGE_TYPE = False
CAN_CHANGE_SIZE = False
DDL_ROLLBACK = True
NEED_GENERATOR = True

FROM = '"%s" AS %s'
LEFT_OUTER_JOIN = 'LEFT OUTER JOIN "%s" AS %s'
FIELD_AS = 'AS'
LIKE = 'ILIKE'
DESC = 'DESC NULLS LAST'

JAM_TYPES = TEXT, INTEGER, FLOAT, CURRENCY, DATE, DATETIME, BOOLEAN, LONGTEXT, KEYS, FILE, IMAGE = range(1, 12)
FIELD_TYPES = {
    INTEGER: 'INTEGER',
    TEXT: 'VARCHAR',
    FLOAT: 'NUMERIC',
    CURRENCY: 'NUMERIC',
    DATE: 'DATE',
    DATETIME: 'TIMESTAMP',
    BOOLEAN: 'INTEGER',
    LONGTEXT: 'TEXT',
    KEYS: 'TEXT',
    FILE: 'TEXT',
    IMAGE: 'TEXT'
}

def connect(database, user, password, host, port, encoding, server):
    return psycopg2.connect(dbname=database, user=user, password=password, host=host, port=port)

get_lastrowid = None

def get_select(query, fields_clause, from_clause, where_clause, group_clause, order_clause, fields):
    start = fields_clause
    end = ''.join([from_clause, where_clause, group_clause, order_clause])
    offset = query['__offset']
    limit = query['__limit']
    result = 'SELECT %s FROM %s' % (start, end)
    if limit:
        result += ' LIMIT %d OFFSET %d' % (limit, offset)
    return result

def process_sql_params(params, cursor):
    result = []
    for p in params:
        if type(p) == tuple:
            value, data_type = p
        else:
            value = p
        result.append(value)
    return result

#~ def process_sql_result(rows):
    #~ return [list(row) for row in rows]

def process_sql_result(rows):
    result = []
    for row in rows:
        fields = []
        for field in row:
            if PY2:
                if type(field) == buffer:
                    field = str(field)
            else:
                if type(field) == memoryview:
                    field = to_unicode(to_bytes(field, 'utf-8'), 'utf-8')
            fields.append(field)
        result.append(fields)
    return result

def cast_date(date_str):
    return "CAST('" + date_str + "' AS DATE)"

def cast_datetime(datetime_str):
    return "CAST('" + datetime_str + "' AS TIMESTAMP)"

def value_literal(index):
    return '%s'

def convert_like(field_name, val, data_type):
    return '%s::text' % field_name, val.upper()

def create_table_sql(table_name, fields, gen_name=None, foreign_fields=None):
    result = []
    primary_key = ''
    seq_name = gen_name
    sql = 'CREATE TABLE "%s"\n(\n' % table_name
    lines = []
    for field in fields:
        line = '"%s" %s' % (field['field_name'], FIELD_TYPES[field['data_type']])
        if field['size'] != 0 and field['data_type'] == TEXT:
            line += '(%d)' % field['size']
        if field['primary_key']:
            primary_key = field['field_name']
            line += ' PRIMARY KEY DEFAULT NEXTVAL(\'"%s"\')' % seq_name
        if field['default_value'] and not field['primary_key']:
            if field['data_type'] == TEXT:
                line += " DEFAULT '%s'" % field['default_value']
            else:
                line += ' DEFAULT %s' % field['default_value']
        lines.append(line)
    sql += ',\n'.join(lines)
    sql += ')\n'
    result.append(sql)
    if primary_key:
        result.insert(0, 'CREATE SEQUENCE "%s"' % seq_name)
        result.append('ALTER SEQUENCE "%s" OWNED BY "%s"."%s"' % (seq_name, table_name, primary_key))
    return result

def delete_table_sql(table_name, gen_name):
    result = []
    result.append('DROP TABLE "%s"' % table_name)
    result.append('DROP SEQUENCE IF EXISTS "%s"' % gen_name)
    return result

def create_index_sql(index_name, table_name, unique, fields, desc):
    return 'CREATE %s INDEX "%s" ON "%s" (%s)' % \
        (unique, index_name, table_name, fields)

def create_foreign_index_sql(table_name, index_name, key, ref, primary_key):
    return 'ALTER TABLE "%s" ADD CONSTRAINT "%s" FOREIGN KEY ("%s") REFERENCES "%s"("%s") MATCH SIMPLE' % \
        (table_name, index_name, key, ref, primary_key)

def delete_index(table_name, index_name):
    return 'DROP INDEX "%s"' % index_name

def delete_foreign_index(table_name, index_name):
    return 'ALTER TABLE "%s" DROP CONSTRAINT "%s"' % (table_name, index_name)

def add_field_sql(table_name, field):
    result = 'ALTER TABLE "%s" ADD COLUMN "%s" %s' % \
        (table_name, field['field_name'], FIELD_TYPES[field['data_type']])
    if field['size']:
        result += '(%d)' % field['size']
    if field['default_value']:
        if field['data_type'] == TEXT:
            result += " DEFAULT '%s'" % field['default_value']
        else:
            result += ' DEFAULT %s' % field['default_value']
    return result

def del_field_sql(table_name, field):
    return 'ALTER TABLE "%s" DROP COLUMN "%s"' % (table_name, field['field_name'])

def change_field_sql(table_name, old_field, new_field):
    result = []
    if FIELD_TYPES[old_field['data_type']] != FIELD_TYPES[new_field['data_type']] \
        or old_field['size'] != new_field['size']:
        raise Exception(u"Don't know how to change field's size or type of %s" % old_field['field_name'])
    if old_field['field_name'] != new_field['field_name']:
        result.append('ALTER TABLE "%s" RENAME COLUMN  "%s" TO "%s"' % \
            (table_name, old_field['field_name'], new_field['field_name']))
    if old_field['default_value'] != new_field['default_value']:
        if new_field['default_value']:
            if new_field['data_type'] == TEXT:
                sql = 'ALTER TABLE "%s" ALTER "%s" SET DEFAULT' % \
                    (table_name, new_field['field_name'])
                sql +=  " '%s'" % new_field['default_value']
            else:
                sql = 'ALTER TABLE "%s" ALTER "%s" SET DEFAULT %s' % \
                    (table_name, new_field['field_name'], new_field['default_value'])
        else:
            sql = 'ALTER TABLE "%s" alter "%s" DROP DEFAULT' % \
                (table_name, new_field['field_name'])
        result.append(sql)
    return result

def next_sequence_value_sql(gen_name):
    return 'SELECT NEXTVAL(\'"%s"\')' % gen_name

def restart_sequence_sql(gen_name, value):
    return 'ALTER SEQUENCE "%s" RESTART WITH %d' % (gen_name, value)

def identifier_case(name):
    return name.lower()

def get_table_names(connection):
    cursor = connection.cursor()
    cursor.execute("SELECT * FROM pg_catalog.pg_tables WHERE schemaname != 'pg_catalog' AND schemaname != 'information_schema'")
    result = cursor.fetchall()
    return [r[1] for r in result]

def get_table_info(connection, table_name, db_name):
    cursor = connection.cursor()
    sql = "select column_name, data_type, character_maximum_length, column_default from information_schema.columns where table_name='%s'" % table_name
    cursor.execute(sql)
    result = cursor.fetchall()
    fields = []
    for column_name, data_type, character_maximum_length, column_default in result:
        try:
            if data_type == 'character varying':
                data_type = 'varchar'
            size = 0
            if character_maximum_length:
                size = character_maximum_length
            default_value = None
            if column_default:
                default_value = column_default.split('::')[0]
                if default_value.find('nextval') != -1:
                    default_value = None
        except:
            pass
        pk = False
        fields.append({
            'field_name': column_name,
            'data_type': data_type.upper(),
            'size': size,
            'default_value': column_default,
            'pk': pk
        })
    return {'fields': fields, 'field_types': FIELD_TYPES}




import os
import sqlite3
import json
import datetime
from shutil import copyfile

from werkzeug._compat import iteritems, to_bytes, to_unicode
from jam.third_party.filelock import FileLock
import jam

LANG_FIELDS = ['id', 'f_name', 'f_language', 'f_country', 'f_abr', 'f_rtl']
LOCALE_FIELDS = [
    'f_decimal_point', 'f_mon_decimal_point',
    'f_mon_thousands_sep', 'f_currency_symbol', 'f_frac_digits', 'f_p_cs_precedes',
    'f_n_cs_precedes', 'f_p_sep_by_space', 'f_n_sep_by_space', 'f_positive_sign',
    'f_negative_sign', 'f_p_sign_posn', 'f_n_sign_posn', 'f_d_fmt', 'f_d_t_fmt'
    ]

FIELDS = LANG_FIELDS + LOCALE_FIELDS

def lang_con(task):
    return sqlite3.connect(os.path.join(task.work_dir, 'langs.sqlite'))

def execute(task, sql, params=None):
    result = None
    con = lang_con(task)
    try:
        cursor = con.cursor()
        if params:
            cursor.execute(sql, params)
        else:
            cursor.execute(sql)
        con.commit()
    except Exception as e:
        print(sql)
        raise Exception(e)
    finally:
        con.close()

def select(task, sql):
    result = None
    con = lang_con(task)
    try:
        cursor = con.cursor()
        cursor.execute(sql)
        result = cursor.fetchall()
        con.rollback()
    except Exception as e:
        print(sql)
        raise Exception(e)
    finally:
        con.close()
    return result

def copy_table(cursor, name):
    cursor.execute('DROP TABLE IF EXISTS SYS_%s' % name)
    cursor.execute("SELECT sql FROM LANGS.sqlite_master WHERE type='table' AND name='JAM_%s'" % name)
    sql = cursor.fetchone()[0]
    cursor.execute(sql.replace('JAM_%s' % name, 'SYS_%s' % name))
    cursor.execute('INSERT INTO SYS_%s SELECT * FROM LANGS.JAM_%s' % (name, name))

def update_langs(task):
    with task.lock('$langs'):
        con = task.create_connection()
        try:
            cursor = con.cursor()
            try:
                cursor.execute('ALTER TABLE SYS_PARAMS ADD COLUMN F_JAM_VERSION TEXT')
            except:
                pass
            cursor.execute('SELECT F_JAM_VERSION, F_LANGUAGE FROM SYS_PARAMS')
            res = cursor.fetchall()
            version = res[0][0]
            language = res[0][1]
            langs_path = os.path.join(task.work_dir, 'langs.sqlite')
            if version != task.app.jam_version or not os.path.exists(langs_path):
                # ~ task.log.info('Version changed!')
                copyfile(os.path.join(os.path.dirname(jam.__file__), 'langs.sqlite'), langs_path)
                os.chmod(os.path.join(task.work_dir, 'langs.sqlite'), 0o666)
                cursor.execute('SELECT ID, F_NAME FROM SYS_LANGS')
                langs = cursor.fetchall()
                langs_list = []
                langs_dict = {}
                for l in langs:
                    langs_list.append(l[1])
                    langs_dict[l[1]] = l[0]
                res = select(task, 'SELECT %s FROM JAM_LANGS ORDER BY ID' % ', '.join(FIELDS))
                for r in res:
                    if langs_dict.get(r[1]):
                        del langs_dict[r[1]]
                    if not r[1] in langs_list:
                        fields = ['DELETED']
                        values = ['?']
                        field_values = [0]
                        for i, value in enumerate(r):
                            if i > 0:
                                fields.append(FIELDS[i])
                                values.append('?')
                                field_values.append(value)
                        sql = "INSERT INTO SYS_LANGS (%s) VALUES (%s)" % (','.join(fields), ','.join(values))
                        cursor.execute(sql, (field_values))
                del_langs = list(langs_dict.values())
                if len(del_langs):
                    if language in del_langs:
                        language = 1
                    sql = "DELETE FROM SYS_LANGS WHERE ID IN (%s)" % ','.join([str(d) for d in del_langs])
                    cursor.execute(sql)
                if language is None:
                    language = 'NULL'
                cursor.execute("UPDATE SYS_PARAMS SET F_JAM_VERSION='%s', F_LANGUAGE=%s" % (task.app.jam_version, language))
                con.commit()
        finally:
            con.close()

def init_locale():
    import locale
    result = {}
    try:
        locale.setlocale(locale.LC_ALL, '')
        loc = locale.localeconv()
        for field in LOCALE_FIELDS:
            setting = field[2:]
            try:
                result[field] = to_unicode(loc[setting], 'utf-8')
            except:
                result[field] = jam.common.DEFAULT_LOCALE[setting.upper()]
    except:
        pass
    try:
        result['f_d_fmt'] = locale.nl_langinfo(locale.D_FMT)
    except:
        result['f_d_fmt'] = '%Y-%m-%d'
    result['f_d_t_fmt'] = '%s %s' % (result['f_d_fmt'], '%H:%M')
    return result

def get_lang_dict(task, language):
    res = select(task, '''
        SELECT K.F_KEYWORD,
            CASE WHEN TRIM(V1.F_VALUE) <> ''
                THEN V1.F_VALUE
                ELSE V2.F_VALUE
            END
        FROM JAM_LANG_KEYS AS K
        LEFT OUTER JOIN JAM_LANG_VALUES AS V1 ON (K.ID = V1.F_KEY AND V1.F_LANG = %s)
        LEFT OUTER JOIN JAM_LANG_VALUES AS V2 ON (K.ID = V2.F_KEY AND V2.F_LANG = %s)
    ''' % (language, 1))
    result = {}
    for key, value in res:
        result[key] = value
    return result

def get_locale_dict(task, language):
    result = {}
    con = task.create_connection()
    try:
        cursor = con.cursor()
        cursor.execute('SELECT %s FROM SYS_LANGS WHERE ID=%s' % (', '.join(LOCALE_FIELDS), language))
        res = cursor.fetchall()
        if len(res):
            for i, field in enumerate(LOCALE_FIELDS):
                result[field[2:].upper()] = res[0][i]
        else:
            raise Exception('Language with id %s is not found' % language)
        con.rollback()
    except:
        result = jam.common.DEFAULT_LOCALE
    finally:
        con.close()
    return result

def get_translation(task, lang1, lang2):
    res = select(task, '''
        SELECT K.ID, K.F_KEYWORD, V1.F_VALUE, V2.F_VALUE
        FROM JAM_LANG_KEYS AS K
        LEFT OUTER JOIN JAM_LANG_VALUES AS V1 ON (K.ID = V1.F_KEY AND V1.F_LANG = %s)
        LEFT OUTER JOIN JAM_LANG_VALUES AS V2 ON (K.ID = V2.F_KEY AND V2.F_LANG = %s)
    ''' % (lang1, lang2))
    return res

def add_lang(task, lang_id, language, country, name, abr, rtl, copy_lang):
    con = lang_con(task)
    try:
        cursor = con.cursor()
        locale = init_locale()
        fields = []
        values = []
        field_values = []
        for key, value in iteritems(locale):
            fields.append(key)
            values.append('?')
            field_values.append(to_unicode(value, 'utf-8'))
        cursor.execute("INSERT INTO JAM_LANGS (ID, F_LANGUAGE, F_COUNTRY, F_NAME, F_ABR, F_RTL, %s) VALUES (?,?,?,?,?,?,%s)" % (','.join(fields), ','.join(values)),
            ([lang_id, language, country, name, abr, rtl] + field_values))
        if copy_lang:
            cursor.execute('''
                SELECT JAM_LANG_KEYS.ID, F_VALUE
                FROM JAM_LANG_VALUES LEFT OUTER JOIN JAM_LANG_KEYS ON JAM_LANG_KEYS.ID = JAM_LANG_VALUES.F_KEY
                WHERE F_LANG = %s
            ''' % copy_lang)
            res = cursor.fetchall()
            recs = []
            for key_id, value in res:
                recs.append((key_id, lang_id, value))
            cursor.executemany("INSERT INTO JAM_LANG_VALUES(F_KEY, F_LANG, F_VALUE) VALUES (?,?,?)", recs)
        con.commit()
        langs = task.sys_langs.copy()
        langs.set_where(id=lang_id)
        langs.open()
        if langs.record_count():
            langs.edit()
            for key, value in iteritems(locale):
                langs.field_by_name(key).value = to_unicode(value, 'utf-8')
            langs.post()
            langs.apply()
    finally:
        con.close()

def save_lang_field(task, lang_id, field_name, value):
    execute(task, 'UPDATE JAM_LANGS SET %s=? WHERE ID=%s' % (field_name, lang_id), (value,))
    con = task.create_connection()
    try:
        cursor = con.cursor()
        cursor.execute('UPDATE SYS_LANGS SET %s=? WHERE ID=%s' % (field_name, lang_id), (value,))
        con.commit()
    finally:
        con.close()
    if task.language == lang_id:
        task.update_lang(lang_id)

def save_translation(task, lang_id, key_id, value):
    res = select(task, 'SELECT ID FROM JAM_LANG_VALUES WHERE F_LANG=%s AND F_KEY=%s' % (lang_id, key_id))
    if len(res):
        execute(task, 'UPDATE JAM_LANG_VALUES SET F_VALUE=? WHERE ID=%s' % (res[0][0]), (value,))
    else:
        execute(task, 'INSERT INTO JAM_LANG_VALUES (F_LANG, F_KEY, F_VALUE) VALUES (?, ?, ?)', (lang_id, key_id, value))

def add_key(task, key):
    result = ''
    con = lang_con(task)
    try:
        cursor = con.cursor()
        cursor.execute("SELECT ID FROM JAM_LANG_KEYS WHERE F_KEYWORD='%s'" % key)
        res = cursor.fetchall()
        if len(res):
            result = 'Keyword exists'
        else:
            cursor.execute('INSERT INTO JAM_LANG_KEYS (F_KEYWORD) VALUES (?)', (key,))
        con.commit()
    finally:
        con.close()
    return result

def del_key(task, key_id):
    result = False
    con = lang_con(task)
    try:
        cursor = con.cursor()
        cursor.execute("DELETE FROM JAM_LANG_VALUES WHERE F_KEY=%s" % key_id)
        cursor.execute("DELETE FROM JAM_LANG_KEYS WHERE ID=%s" % key_id)
        con.commit()
        result = True
    finally:
        con.close()
    return result

def get_dict(task, language):
    res =  select(task, '''
            SELECT JAM_LANG_KEYS.F_KEYWORD, F_VALUE
            FROM JAM_LANG_VALUES LEFT OUTER JOIN JAM_LANG_KEYS ON JAM_LANG_KEYS.ID = JAM_LANG_VALUES.F_KEY
            WHERE F_LANG = %s
        ''' % language)
    result = {}
    for key, value in res:
        result[key] = value
    return result

def export_lang(task, lang_id, host):
    names = FIELDS[1:]
    lang = select(task, 'SELECT %s FROM JAM_LANGS WHERE ID=%s' % (', '.join(names), lang_id))
    if len(lang):
        language = {}
        for i in range(len(lang[0])):
            language[names[i]] = lang[0][i]

        translation = get_dict(task, lang_id)
        content = json.dumps({'language': language, 'translation': translation})

        name = language['f_name'].replace(' ', '_')
        file_name = '%s_%s.lang' % (name, datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S"))
        return {'file_name': file_name, 'content': content}

def import_lang(task, file_path):
    error = ''
    try:
        with open(file_path, 'r') as f:
            content = to_unicode(f.read(), 'utf-8')
        content = json.loads(content)
        language = content['language']
        translation = content['translation']

        con = lang_con(task)
        sys_con = task.create_connection()
        try:
            cursor = con.cursor()
            cursor.execute('SELECT ID FROM JAM_LANGS WHERE F_LANGUAGE=%s AND F_COUNTRY=%s' % (language['f_language'], language['f_country']))
            res = cursor.fetchall()
            if len(res):
                lang_id = res[0][0]
                fields = []
                field_values = []
                for key, value in iteritems(language):
                    fields.append('%s=?' % key)
                    field_values.append(value)
                fields = ',' .join(fields)
                cursor.execute("UPDATE JAM_LANGS SET %s WHERE ID=%s" % (fields, lang_id), field_values)

                sys_cursor = sys_con.cursor()
                sys_cursor.execute("UPDATE SYS_LANGS SET %s WHERE ID=%s" % (fields, lang_id), field_values)
                sys_con.commit()
            else:
                fields = []
                values = []
                field_values = []
                for key, value in iteritems(language):
                    fields.append(key)
                    field_values.append(value)
                    values.append('?')
                cursor.execute('INSERT INTO JAM_LANGS (%s) VALUES (%s)' % (','.join(fields), ','.join(values)), field_values)
                cursor.execute('SELECT ID FROM JAM_LANGS WHERE F_LANGUAGE=%s AND F_COUNTRY=%s' % (language['f_language'], language['f_country']))
                res = cursor.fetchall()
                lang_id = res[0][0]
                fields.append('DELETED')
                values.append('?')
                field_values.append(0)
                sys_cursor = sys_con.cursor()
                sys_cursor.execute('INSERT INTO SYS_LANGS (%s) VALUES (%s)' % (','.join(fields), ','.join(values)), field_values)
                sys_con.commit()
            if lang_id:
                cursor.execute('SELECT ID, F_KEYWORD FROM JAM_LANG_KEYS')
                res = cursor.fetchall()
                keys = {}
                for r in res:
                    keys[r[1]] = r[0]
                recs = []
                for keyword, value in iteritems(translation):
                    key_id = keys.get(keyword)
                    if key_id:
                        cursor.execute('SELECT ID FROM JAM_LANG_VALUES WHERE F_LANG=%s AND F_KEY=%s' % (lang_id, key_id))
                        res = cursor.fetchall()
                        if len(res):
                            cursor.execute('UPDATE JAM_LANG_VALUES SET F_VALUE=? WHERE ID=%s' % (res[0][0]), (value,))
                        else:
                            cursor.execute('INSERT INTO JAM_LANG_VALUES (F_LANG, F_KEY, F_VALUE) VALUES (?, ?, ?)', (lang_id, key_id, value))
            con.commit()
        finally:
            con.close()
            sys_con.close()
    except Exception as e:
        print(e)
        error = 'Can not import language'



from collections import deque
from collections import namedtuple
import operator

from . import operators
from .visitors import ExtendedInternalTraversal
from .visitors import InternalTraversal
from .. import util
from ..inspection import inspect

SKIP_TRAVERSE = util.symbol("skip_traverse")
COMPARE_FAILED = False
COMPARE_SUCCEEDED = True
NO_CACHE = util.symbol("no_cache")
CACHE_IN_PLACE = util.symbol("cache_in_place")
CALL_GEN_CACHE_KEY = util.symbol("call_gen_cache_key")
STATIC_CACHE_KEY = util.symbol("static_cache_key")


def compare(obj1, obj2, **kw):
    if kw.get("use_proxies", False):
        strategy = ColIdentityComparatorStrategy()
    else:
        strategy = TraversalComparatorStrategy()

    return strategy.compare(obj1, obj2, **kw)


class HasCacheKey(object):
    _cache_key_traversal = NO_CACHE

    __slots__ = ()

    def _gen_cache_key(self, anon_map, bindparams):
        """return an optional cache key.

        The cache key is a tuple which can contain any series of
        objects that are hashable and also identifies
        this object uniquely within the presence of a larger SQL expression
        or statement, for the purposes of caching the resulting query.

        The cache key should be based on the SQL compiled structure that would
        ultimately be produced.   That is, two structures that are composed in
        exactly the same way should produce the same cache key; any difference
        in the structures that would affect the SQL string or the type handlers
        should result in a different cache key.

        If a structure cannot produce a useful cache key, it should raise
        NotImplementedError, which will result in the entire structure
        for which it's part of not being useful as a cache key.


        """

        idself = id(self)

        if anon_map is not None:
            if idself in anon_map:
                return (anon_map[idself], self.__class__)
            else:
                # inline of
                # id_ = anon_map[idself]
                anon_map[idself] = id_ = str(anon_map.index)
                anon_map.index += 1
        else:
            id_ = None

        _cache_key_traversal = self._cache_key_traversal
        if _cache_key_traversal is None:
            try:
                _cache_key_traversal = self._traverse_internals
            except AttributeError:
                _cache_key_traversal = NO_CACHE

        if _cache_key_traversal is NO_CACHE:
            if anon_map is not None:
                anon_map[NO_CACHE] = True
            return None

        result = (id_, self.__class__)

        # inline of _cache_key_traversal_visitor.run_generated_dispatch()
        try:
            dispatcher = self.__class__.__dict__[
                "_generated_cache_key_traversal"
            ]
        except KeyError:
            dispatcher = _cache_key_traversal_visitor.generate_dispatch(
                self, _cache_key_traversal, "_generated_cache_key_traversal"
            )

        for attrname, obj, meth in dispatcher(
            self, _cache_key_traversal_visitor
        ):
            if obj is not None:
                if meth is CACHE_IN_PLACE:
                    # cache in place is always going to be a Python
                    # tuple, dict, list, etc. so we can do a boolean check
                    if obj:
                        result += (attrname, obj)
                elif meth is STATIC_CACHE_KEY:
                    result += (attrname, obj._static_cache_key)
                elif meth is CALL_GEN_CACHE_KEY:
                    result += (
                        attrname,
                        obj._gen_cache_key(anon_map, bindparams),
                    )
                elif meth is InternalTraversal.dp_clauseelement_list:
                    if obj:
                        result += (
                            attrname,
                            tuple(
                                [
                                    elem._gen_cache_key(anon_map, bindparams)
                                    for elem in obj
                                ]
                            ),
                        )
                else:
                    # note that all the "ClauseElement" standalone cases
                    # here have been handled by inlines above; so we can
                    # safely assume the object is a standard list/tuple/dict
                    # which we can skip if it evaluates to false.
                    # improvement would be to have this as a flag delivered
                    # up front in the dispatcher list
                    if obj:
                        result += meth(
                            attrname, obj, self, anon_map, bindparams
                        )

        return result

    def _generate_cache_key(self):
        """return a cache key.

        The cache key is a tuple which can contain any series of
        objects that are hashable and also identifies
        this object uniquely within the presence of a larger SQL expression
        or statement, for the purposes of caching the resulting query.

        The cache key should be based on the SQL compiled structure that would
        ultimately be produced.   That is, two structures that are composed in
        exactly the same way should produce the same cache key; any difference
        in the structures that would affect the SQL string or the type handlers
        should result in a different cache key.

        The cache key returned by this method is an instance of
        :class:`.CacheKey`, which consists of a tuple representing the
        cache key, as well as a list of :class:`.BindParameter` objects
        which are extracted from the expression.   While two expressions
        that produce identical cache key tuples will themselves generate
        identical SQL strings, the list of :class:`.BindParameter` objects
        indicates the bound values which may have different values in
        each one; these bound parameters must be consulted in order to
        execute the statement with the correct parameters.

        a :class:`.ClauseElement` structure that does not implement
        a :meth:`._gen_cache_key` method and does not implement a
        :attr:`.traverse_internals` attribute will not be cacheable; when
        such an element is embedded into a larger structure, this method
        will return None, indicating no cache key is available.

        """
        bindparams = []

        _anon_map = anon_map()
        key = self._gen_cache_key(_anon_map, bindparams)
        if NO_CACHE in _anon_map:
            return None
        else:
            return CacheKey(key, bindparams)


class CacheKey(namedtuple("CacheKey", ["key", "bindparams"])):
    def __hash__(self):
        return hash(self.key)

    def __eq__(self, other):
        return self.key == other.key


def _clone(element, **kw):
    return element._clone()


class _CacheKey(ExtendedInternalTraversal):
    # very common elements are inlined into the main _get_cache_key() method
    # to produce a dramatic savings in Python function call overhead

    visit_has_cache_key = visit_clauseelement = CALL_GEN_CACHE_KEY
    visit_clauseelement_list = InternalTraversal.dp_clauseelement_list
    visit_string = (
        visit_boolean
    ) = visit_operator = visit_plain_obj = CACHE_IN_PLACE
    visit_statement_hint_list = CACHE_IN_PLACE
    visit_type = STATIC_CACHE_KEY

    def visit_inspectable(self, attrname, obj, parent, anon_map, bindparams):
        return self.visit_has_cache_key(
            attrname, inspect(obj), parent, anon_map, bindparams
        )

    def visit_multi(self, attrname, obj, parent, anon_map, bindparams):
        return (
            attrname,
            obj._gen_cache_key(anon_map, bindparams)
            if isinstance(obj, HasCacheKey)
            else obj,
        )

    def visit_multi_list(self, attrname, obj, parent, anon_map, bindparams):
        return (
            attrname,
            tuple(
                elem._gen_cache_key(anon_map, bindparams)
                if isinstance(elem, HasCacheKey)
                else elem
                for elem in obj
            ),
        )

    def visit_has_cache_key_tuples(
        self, attrname, obj, parent, anon_map, bindparams
    ):
        if not obj:
            return ()
        return (
            attrname,
            tuple(
                tuple(
                    elem._gen_cache_key(anon_map, bindparams)
                    for elem in tup_elem
                )
                for tup_elem in obj
            ),
        )

    def visit_has_cache_key_list(
        self, attrname, obj, parent, anon_map, bindparams
    ):
        if not obj:
            return ()
        return (
            attrname,
            tuple(elem._gen_cache_key(anon_map, bindparams) for elem in obj),
        )

    def visit_inspectable_list(
        self, attrname, obj, parent, anon_map, bindparams
    ):
        return self.visit_has_cache_key_list(
            attrname, [inspect(o) for o in obj], parent, anon_map, bindparams
        )

    def visit_clauseelement_tuples(
        self, attrname, obj, parent, anon_map, bindparams
    ):
        return self.visit_has_cache_key_tuples(
            attrname, obj, parent, anon_map, bindparams
        )

    def visit_anon_name(self, attrname, obj, parent, anon_map, bindparams):
        from . import elements

        name = obj
        if isinstance(name, elements._anonymous_label):
            name = name.apply_map(anon_map)

        return (attrname, name)

    def visit_fromclause_ordered_set(
        self, attrname, obj, parent, anon_map, bindparams
    ):
        if not obj:
            return ()
        return (
            attrname,
            tuple([elem._gen_cache_key(anon_map, bindparams) for elem in obj]),
        )

    def visit_clauseelement_unordered_set(
        self, attrname, obj, parent, anon_map, bindparams
    ):
        if not obj:
            return ()
        cache_keys = [
            elem._gen_cache_key(anon_map, bindparams) for elem in obj
        ]
        return (
            attrname,
            tuple(
                sorted(cache_keys)
            ),  # cache keys all start with (id_, class)
        )

    def visit_named_ddl_element(
        self, attrname, obj, parent, anon_map, bindparams
    ):
        return (attrname, obj.name)

    def visit_prefix_sequence(
        self, attrname, obj, parent, anon_map, bindparams
    ):
        if not obj:
            return ()
        return (
            attrname,
            tuple(
                [
                    (clause._gen_cache_key(anon_map, bindparams), strval)
                    for clause, strval in obj
                ]
            ),
        )

    def visit_table_hint_list(
        self, attrname, obj, parent, anon_map, bindparams
    ):
        if not obj:
            return ()

        return (
            attrname,
            tuple(
                [
                    (
                        clause._gen_cache_key(anon_map, bindparams),
                        dialect_name,
                        text,
                    )
                    for (clause, dialect_name), text in obj.items()
                ]
            ),
        )

    def visit_plain_dict(self, attrname, obj, parent, anon_map, bindparams):
        return (attrname, tuple([(key, obj[key]) for key in sorted(obj)]))

    def visit_string_clauseelement_dict(
        self, attrname, obj, parent, anon_map, bindparams
    ):
        return (
            attrname,
            tuple(
                (key, obj[key]._gen_cache_key(anon_map, bindparams))
                for key in sorted(obj)
            ),
        )

    def visit_string_multi_dict(
        self, attrname, obj, parent, anon_map, bindparams
    ):
        return (
            attrname,
            tuple(
                (
                    key,
                    value._gen_cache_key(anon_map, bindparams)
                    if isinstance(value, HasCacheKey)
                    else value,
                )
                for key, value in [(key, obj[key]) for key in sorted(obj)]
            ),
        )

    def visit_fromclause_canonical_column_collection(
        self, attrname, obj, parent, anon_map, bindparams
    ):
        return (
            attrname,
            tuple(col._gen_cache_key(anon_map, bindparams) for col in obj),
        )

    def visit_unknown_structure(
        self, attrname, obj, parent, anon_map, bindparams
    ):
        anon_map[NO_CACHE] = True
        return ()


_cache_key_traversal_visitor = _CacheKey()


class _CopyInternals(InternalTraversal):
    """Generate a _copy_internals internal traversal dispatch for classes
    with a _traverse_internals collection."""

    def visit_clauseelement(self, parent, element, clone=_clone, **kw):
        return clone(element, **kw)

    def visit_clauseelement_list(self, parent, element, clone=_clone, **kw):
        return [clone(clause, **kw) for clause in element]

    def visit_clauseelement_tuples(self, parent, element, clone=_clone, **kw):
        return [
            tuple(clone(tup_elem, **kw) for tup_elem in elem)
            for elem in element
        ]

    def visit_string_clauseelement_dict(
        self, parent, element, clone=_clone, **kw
    ):
        return dict(
            (key, clone(value, **kw)) for key, value in element.items()
        )


_copy_internals = _CopyInternals()


class _GetChildren(InternalTraversal):
    """Generate a _children_traversal internal traversal dispatch for classes
    with a _traverse_internals collection."""

    def visit_has_cache_key(self, element, **kw):
        return (element,)

    def visit_clauseelement(self, element, **kw):
        return (element,)

    def visit_clauseelement_list(self, element, **kw):
        return tuple(element)

    def visit_clauseelement_tuples(self, element, **kw):
        tup = ()
        for elem in element:
            tup += elem
        return tup

    def visit_fromclause_canonical_column_collection(self, element, **kw):
        if kw.get("column_collections", False):
            return tuple(element)
        else:
            return ()

    def visit_string_clauseelement_dict(self, element, **kw):
        return tuple(element.values())

    def visit_fromclause_ordered_set(self, element, **kw):
        return tuple(element)

    def visit_clauseelement_unordered_set(self, element, **kw):
        return tuple(element)


_get_children = _GetChildren()


@util.dependencies("sqlalchemy.sql.elements")
def _resolve_name_for_compare(elements, element, name, anon_map, **kw):
    if isinstance(name, elements._anonymous_label):
        name = name.apply_map(anon_map)

    return name


class anon_map(dict):
    """A map that creates new keys for missing key access.

    Produces an incrementing sequence given a series of unique keys.

    This is similar to the compiler prefix_anon_map class although simpler.

    Inlines the approach taken by :class:`sqlalchemy.util.PopulateDict` which
    is otherwise usually used for this type of operation.

    """

    def __init__(self):
        self.index = 0

    def __missing__(self, key):
        self[key] = val = str(self.index)
        self.index += 1
        return val


class TraversalComparatorStrategy(InternalTraversal, util.MemoizedSlots):
    __slots__ = "stack", "cache", "anon_map"

    def __init__(self):
        self.stack = deque()
        self.cache = set()

    def _memoized_attr_anon_map(self):
        return (anon_map(), anon_map())

    def compare(self, obj1, obj2, **kw):
        stack = self.stack
        cache = self.cache

        compare_annotations = kw.get("compare_annotations", False)

        stack.append((obj1, obj2))

        while stack:
            left, right = stack.popleft()

            if left is right:
                continue
            elif left is None or right is None:
                # we know they are different so no match
                return False
            elif (left, right) in cache:
                continue
            cache.add((left, right))

            visit_name = left.__visit_name__
            if visit_name != right.__visit_name__:
                return False

            meth = getattr(self, "compare_%s" % visit_name, None)

            if meth:
                attributes_compared = meth(left, right, **kw)
                if attributes_compared is COMPARE_FAILED:
                    return False
                elif attributes_compared is SKIP_TRAVERSE:
                    continue

                # attributes_compared is returned as a list of attribute
                # names that were "handled" by the comparison method above.
                # remaining attribute names in the _traverse_internals
                # will be compared.
            else:
                attributes_compared = ()

            for (
                (left_attrname, left_visit_sym),
                (right_attrname, right_visit_sym),
            ) in util.zip_longest(
                left._traverse_internals,
                right._traverse_internals,
                fillvalue=(None, None),
            ):
                if not compare_annotations and (
                    (left_attrname == "_annotations_cache_key")
                    or (right_attrname == "_annotations_cache_key")
                ):
                    continue

                if (
                    left_attrname != right_attrname
                    or left_visit_sym is not right_visit_sym
                ):
                    return False
                elif left_attrname in attributes_compared:
                    continue

                dispatch = self.dispatch(left_visit_sym)
                left_child = operator.attrgetter(left_attrname)(left)
                right_child = operator.attrgetter(right_attrname)(right)
                if left_child is None:
                    if right_child is not None:
                        return False
                    else:
                        continue

                comparison = dispatch(
                    left, left_child, right, right_child, **kw
                )
                if comparison is COMPARE_FAILED:
                    return False

        return True

    def compare_inner(self, obj1, obj2, **kw):
        comparator = self.__class__()
        return comparator.compare(obj1, obj2, **kw)

    def visit_has_cache_key(
        self, left_parent, left, right_parent, right, **kw
    ):
        if left._gen_cache_key(self.anon_map[0], []) != right._gen_cache_key(
            self.anon_map[1], []
        ):
            return COMPARE_FAILED

    def visit_clauseelement(
        self, left_parent, left, right_parent, right, **kw
    ):
        self.stack.append((left, right))

    def visit_fromclause_canonical_column_collection(
        self, left_parent, left, right_parent, right, **kw
    ):
        for lcol, rcol in util.zip_longest(left, right, fillvalue=None):
            self.stack.append((lcol, rcol))

    def visit_fromclause_derived_column_collection(
        self, left_parent, left, right_parent, right, **kw
    ):
        pass

    def visit_string_clauseelement_dict(
        self, left_parent, left, right_parent, right, **kw
    ):
        for lstr, rstr in util.zip_longest(
            sorted(left), sorted(right), fillvalue=None
        ):
            if lstr != rstr:
                return COMPARE_FAILED
            self.stack.append((left[lstr], right[rstr]))

    def visit_clauseelement_tuples(
        self, left_parent, left, right_parent, right, **kw
    ):
        for ltup, rtup in util.zip_longest(left, right, fillvalue=None):
            if ltup is None or rtup is None:
                return COMPARE_FAILED

            for l, r in util.zip_longest(ltup, rtup, fillvalue=None):
                self.stack.append((l, r))

    def visit_clauseelement_list(
        self, left_parent, left, right_parent, right, **kw
    ):
        for l, r in util.zip_longest(left, right, fillvalue=None):
            self.stack.append((l, r))

    def _compare_unordered_sequences(self, seq1, seq2, **kw):
        if seq1 is None:
            return seq2 is None

        completed = set()
        for clause in seq1:
            for other_clause in set(seq2).difference(completed):
                if self.compare_inner(clause, other_clause, **kw):
                    completed.add(other_clause)
                    break
        return len(completed) == len(seq1) == len(seq2)

    def visit_clauseelement_unordered_set(
        self, left_parent, left, right_parent, right, **kw
    ):
        return self._compare_unordered_sequences(left, right, **kw)

    def visit_fromclause_ordered_set(
        self, left_parent, left, right_parent, right, **kw
    ):
        for l, r in util.zip_longest(left, right, fillvalue=None):
            self.stack.append((l, r))

    def visit_string(self, left_parent, left, right_parent, right, **kw):
        return left == right

    def visit_anon_name(self, left_parent, left, right_parent, right, **kw):
        return _resolve_name_for_compare(
            left_parent, left, self.anon_map[0], **kw
        ) == _resolve_name_for_compare(
            right_parent, right, self.anon_map[1], **kw
        )

    def visit_boolean(self, left_parent, left, right_parent, right, **kw):
        return left == right

    def visit_operator(self, left_parent, left, right_parent, right, **kw):
        return left is right

    def visit_type(self, left_parent, left, right_parent, right, **kw):
        return left._compare_type_affinity(right)

    def visit_plain_dict(self, left_parent, left, right_parent, right, **kw):
        return left == right

    def visit_plain_obj(self, left_parent, left, right_parent, right, **kw):
        return left == right

    def visit_named_ddl_element(
        self, left_parent, left, right_parent, right, **kw
    ):
        if left is None:
            if right is not None:
                return COMPARE_FAILED

        return left.name == right.name

    def visit_prefix_sequence(
        self, left_parent, left, right_parent, right, **kw
    ):
        for (l_clause, l_str), (r_clause, r_str) in util.zip_longest(
            left, right, fillvalue=(None, None)
        ):
            if l_str != r_str:
                return COMPARE_FAILED
            else:
                self.stack.append((l_clause, r_clause))

    def visit_table_hint_list(
        self, left_parent, left, right_parent, right, **kw
    ):
        left_keys = sorted(left, key=lambda elem: (elem[0].fullname, elem[1]))
        right_keys = sorted(
            right, key=lambda elem: (elem[0].fullname, elem[1])
        )
        for (ltable, ldialect), (rtable, rdialect) in util.zip_longest(
            left_keys, right_keys, fillvalue=(None, None)
        ):
            if ldialect != rdialect:
                return COMPARE_FAILED
            elif left[(ltable, ldialect)] != right[(rtable, rdialect)]:
                return COMPARE_FAILED
            else:
                self.stack.append((ltable, rtable))

    def visit_statement_hint_list(
        self, left_parent, left, right_parent, right, **kw
    ):
        return left == right

    def visit_unknown_structure(
        self, left_parent, left, right_parent, right, **kw
    ):
        raise NotImplementedError()

    def compare_clauselist(self, left, right, **kw):
        if left.operator is right.operator:
            if operators.is_associative(left.operator):
                if self._compare_unordered_sequences(
                    left.clauses, right.clauses, **kw
                ):
                    return ["operator", "clauses"]
                else:
                    return COMPARE_FAILED
            else:
                return ["operator"]
        else:
            return COMPARE_FAILED

    def compare_binary(self, left, right, **kw):
        if left.operator == right.operator:
            if operators.is_commutative(left.operator):
                if (
                    compare(left.left, right.left, **kw)
                    and compare(left.right, right.right, **kw)
                ) or (
                    compare(left.left, right.right, **kw)
                    and compare(left.right, right.left, **kw)
                ):
                    return ["operator", "negate", "left", "right"]
                else:
                    return COMPARE_FAILED
            else:
                return ["operator", "negate"]
        else:
            return COMPARE_FAILED

    def compare_bindparam(self, left, right, **kw):
        compare_values = kw.pop("compare_values", True)
        if compare_values:
            return []
        else:
            # this means, "skip these, we already compared"
            return ["callable", "value"]


class ColIdentityComparatorStrategy(TraversalComparatorStrategy):
    def compare_column_element(
        self, left, right, use_proxies=True, equivalents=(), **kw
    ):
        """Compare ColumnElements using proxies and equivalent collections.

        This is a comparison strategy specific to the ORM.
        """

        to_compare = (right,)
        if equivalents and right in equivalents:
            to_compare = equivalents[right].union(to_compare)

        for oth in to_compare:
            if use_proxies and left.shares_lineage(oth):
                return SKIP_TRAVERSE
            elif hash(left) == hash(right):
                return SKIP_TRAVERSE
        else:
            return COMPARE_FAILED

    def compare_column(self, left, right, **kw):
        return self.compare_column_element(left, right, **kw)

    def compare_label(self, left, right, **kw):
        return self.compare_column_element(left, right, **kw)

    def compare_table(self, left, right, **kw):
        # tables compare on identity, since it's not really feasible to
        # compare them column by column with the above rules
        return SKIP_TRAVERSE if left is right else COMPARE_FAILED



# -*- coding: utf-8 -*-
# Copyright JS Foundation and other contributors, https://js.foundation/
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import unicode_literals

# Generated by generate-xhtml-entities.js. DO NOT MODIFY!

XHTMLEntities = {
    'quot': "\u0022",
    'amp': "\u0026",
    'apos': "\u0027",
    'gt': "\u003E",
    'nbsp': "\u00A0",
    'iexcl': "\u00A1",
    'cent': "\u00A2",
    'pound': "\u00A3",
    'curren': "\u00A4",
    'yen': "\u00A5",
    'brvbar': "\u00A6",
    'sect': "\u00A7",
    'uml': "\u00A8",
    'copy': "\u00A9",
    'ordf': "\u00AA",
    'laquo': "\u00AB",
    'not': "\u00AC",
    'shy': "\u00AD",
    'reg': "\u00AE",
    'macr': "\u00AF",
    'deg': "\u00B0",
    'plusmn': "\u00B1",
    'sup2': "\u00B2",
    'sup3': "\u00B3",
    'acute': "\u00B4",
    'micro': "\u00B5",
    'para': "\u00B6",
    'middot': "\u00B7",
    'cedil': "\u00B8",
    'sup1': "\u00B9",
    'ordm': "\u00BA",
    'raquo': "\u00BB",
    'frac14': "\u00BC",
    'frac12': "\u00BD",
    'frac34': "\u00BE",
    'iquest': "\u00BF",
    'Agrave': "\u00C0",
    'Aacute': "\u00C1",
    'Acirc': "\u00C2",
    'Atilde': "\u00C3",
    'Auml': "\u00C4",
    'Aring': "\u00C5",
    'AElig': "\u00C6",
    'Ccedil': "\u00C7",
    'Egrave': "\u00C8",
    'Eacute': "\u00C9",
    'Ecirc': "\u00CA",
    'Euml': "\u00CB",
    'Igrave': "\u00CC",
    'Iacute': "\u00CD",
    'Icirc': "\u00CE",
    'Iuml': "\u00CF",
    'ETH': "\u00D0",
    'Ntilde': "\u00D1",
    'Ograve': "\u00D2",
    'Oacute': "\u00D3",
    'Ocirc': "\u00D4",
    'Otilde': "\u00D5",
    'Ouml': "\u00D6",
    'times': "\u00D7",
    'Oslash': "\u00D8",
    'Ugrave': "\u00D9",
    'Uacute': "\u00DA",
    'Ucirc': "\u00DB",
    'Uuml': "\u00DC",
    'Yacute': "\u00DD",
    'THORN': "\u00DE",
    'szlig': "\u00DF",
    'agrave': "\u00E0",
    'aacute': "\u00E1",
    'acirc': "\u00E2",
    'atilde': "\u00E3",
    'auml': "\u00E4",
    'aring': "\u00E5",
    'aelig': "\u00E6",
    'ccedil': "\u00E7",
    'egrave': "\u00E8",
    'eacute': "\u00E9",
    'ecirc': "\u00EA",
    'euml': "\u00EB",
    'igrave': "\u00EC",
    'iacute': "\u00ED",
    'icirc': "\u00EE",
    'iuml': "\u00EF",
    'eth': "\u00F0",
    'ntilde': "\u00F1",
    'ograve': "\u00F2",
    'oacute': "\u00F3",
    'ocirc': "\u00F4",
    'otilde': "\u00F5",
    'ouml': "\u00F6",
    'divide': "\u00F7",
    'oslash': "\u00F8",
    'ugrave': "\u00F9",
    'uacute': "\u00FA",
    'ucirc': "\u00FB",
    'uuml': "\u00FC",
    'yacute': "\u00FD",
    'thorn': "\u00FE",
    'yuml': "\u00FF",
    'OElig': "\u0152",
    'oelig': "\u0153",
    'Scaron': "\u0160",
    'scaron': "\u0161",
    'Yuml': "\u0178",
    'fnof': "\u0192",
    'circ': "\u02C6",
    'tilde': "\u02DC",
    'Alpha': "\u0391",
    'Beta': "\u0392",
    'Gamma': "\u0393",
    'Delta': "\u0394",
    'Epsilon': "\u0395",
    'Zeta': "\u0396",
    'Eta': "\u0397",
    'Theta': "\u0398",
    'Iota': "\u0399",
    'Kappa': "\u039A",
    'Lambda': "\u039B",
    'Mu': "\u039C",
    'Nu': "\u039D",
    'Xi': "\u039E",
    'Omicron': "\u039F",
    'Pi': "\u03A0",
    'Rho': "\u03A1",
    'Sigma': "\u03A3",
    'Tau': "\u03A4",
    'Upsilon': "\u03A5",
    'Phi': "\u03A6",
    'Chi': "\u03A7",
    'Psi': "\u03A8",
    'Omega': "\u03A9",
    'alpha': "\u03B1",
    'beta': "\u03B2",
    'gamma': "\u03B3",
    'delta': "\u03B4",
    'epsilon': "\u03B5",
    'zeta': "\u03B6",
    'eta': "\u03B7",
    'theta': "\u03B8",
    'iota': "\u03B9",
    'kappa': "\u03BA",
    'lambda': "\u03BB",
    'mu': "\u03BC",
    'nu': "\u03BD",
    'xi': "\u03BE",
    'omicron': "\u03BF",
    'pi': "\u03C0",
    'rho': "\u03C1",
    'sigmaf': "\u03C2",
    'sigma': "\u03C3",
    'tau': "\u03C4",
    'upsilon': "\u03C5",
    'phi': "\u03C6",
    'chi': "\u03C7",
    'psi': "\u03C8",
    'omega': "\u03C9",
    'thetasym': "\u03D1",
    'upsih': "\u03D2",
    'piv': "\u03D6",
    'ensp': "\u2002",
    'emsp': "\u2003",
    'thinsp': "\u2009",
    'zwnj': "\u200C",
    'zwj': "\u200D",
    'lrm': "\u200E",
    'rlm': "\u200F",
    'ndash': "\u2013",
    'mdash': "\u2014",
    'lsquo': "\u2018",
    'rsquo': "\u2019",
    'sbquo': "\u201A",
    'ldquo': "\u201C",
    'rdquo': "\u201D",
    'bdquo': "\u201E",
    'dagger': "\u2020",
    'Dagger': "\u2021",
    'bull': "\u2022",
    'hellip': "\u2026",
    'permil': "\u2030",
    'prime': "\u2032",
    'Prime': "\u2033",
    'lsaquo': "\u2039",
    'rsaquo': "\u203A",
    'oline': "\u203E",
    'frasl': "\u2044",
    'euro': "\u20AC",
    'image': "\u2111",
    'weierp': "\u2118",
    'real': "\u211C",
    'trade': "\u2122",
    'alefsym': "\u2135",
    'larr': "\u2190",
    'uarr': "\u2191",
    'rarr': "\u2192",
    'darr': "\u2193",
    'harr': "\u2194",
    'crarr': "\u21B5",
    'lArr': "\u21D0",
    'uArr': "\u21D1",
    'rArr': "\u21D2",
    'dArr': "\u21D3",
    'hArr': "\u21D4",
    'forall': "\u2200",
    'part': "\u2202",
    'exist': "\u2203",
    'empty': "\u2205",
    'nabla': "\u2207",
    'isin': "\u2208",
    'notin': "\u2209",
    'ni': "\u220B",
    'prod': "\u220F",
    'sum': "\u2211",
    'minus': "\u2212",
    'lowast': "\u2217",
    'radic': "\u221A",
    'prop': "\u221D",
    'infin': "\u221E",
    'ang': "\u2220",
    'and': "\u2227",
    'or': "\u2228",
    'cap': "\u2229",
    'cup': "\u222A",
    'int': "\u222B",
    'there4': "\u2234",
    'sim': "\u223C",
    'cong': "\u2245",
    'asymp': "\u2248",
    'ne': "\u2260",
    'equiv': "\u2261",
    'le': "\u2264",
    'ge': "\u2265",
    'sub': "\u2282",
    'sup': "\u2283",
    'nsub': "\u2284",
    'sube': "\u2286",
    'supe': "\u2287",
    'oplus': "\u2295",
    'otimes': "\u2297",
    'perp': "\u22A5",
    'sdot': "\u22C5",
    'lceil': "\u2308",
    'rceil': "\u2309",
    'lfloor': "\u230A",
    'rfloor': "\u230B",
    'loz': "\u25CA",
    'spades': "\u2660",
    'clubs': "\u2663",
    'hearts': "\u2665",
    'diams': "\u2666",
    'lang': "\u27E8",
    'rang': "\u27E9",
}



import sys
import os
import json
import uuid
import traceback
import datetime, time
from types import MethodType
import mimetypes
import logging
import jam

sys.path.insert(0, os.path.realpath(os.path.join(os.path.dirname(jam.__file__), 'third_party')))

from werkzeug.wrappers import Request, Response
from werkzeug.routing import Map, Rule
from werkzeug.exceptions import HTTPException, NotFound
from werkzeug.middleware.shared_data import SharedDataMiddleware
from werkzeug.wsgi import peek_path_info, get_path_info
from werkzeug.local import Local, LocalManager
from werkzeug.http import parse_date, http_date
from werkzeug.utils import cached_property, secure_filename
from werkzeug._compat import to_unicode, to_bytes
from werkzeug.secure_cookie.securecookie import SecureCookie

from .third_party.six import get_function_code

from .common import consts, error_message, file_write
from .common import consts, ProjectNotCompleted
from .common import json_defaul_handler, compressBuf
from .common import validate_image, valid_uploaded_file
from .admin.admin import create_admin, login_user, get_roles
from .admin.admin import user_valid_ip, user_valid_uuid
from .admin.builder import update_events_code
from .admin.import_metadata import import_metadata
from .items import AbortException
from .admin.task import create_task, reload_task

class JamSecureCookie(SecureCookie):
    serialization_method = json

    def save_cookie(
        self,
        response,
        key="session",
        expires=None,
        session_expires=None,
        max_age=None,
        path="/",
        domain=None,
        secure=None,
        httponly=False,
        force=False,
        samesite=None
    ):
        if force or self.should_save:
            data = self.serialize(session_expires or expires)
            response.set_cookie(
                key,
                data,
                expires=expires,
                max_age=max_age,
                path=path,
                domain=domain,
                secure=secure,
                httponly=httponly,
                samesite=samesite
            )


class JamRequest(Request):

    @cached_property
    def session_key(self):
        if self.task.app.admin == self.task:
            return '%s_session_%s' % (self.task.item_name, self.environ['SERVER_PORT'])
        else:
            return '%s_session' % (self.task.item_name)

    @cached_property
    def client_cookie(self):
        return JamSecureCookie.load_cookie(self, key=self.session_key, secret_key=self.task.app.admin.secret_key)

    def save_client_cookie(self, response, app, task):
        if task:
            cookie = self.client_cookie
            expires = None
            if consts.SAFE_MODE and task.timeout:
                expires = time.time() + task.timeout
                cookie.modified = True
            cookie.save_cookie(response, key=self.session_key, session_expires=expires, httponly=True, samesite='Lax')

def create_application(from_file=None, load_task=False, testing=False):
    if from_file:
        if os.path.isfile(from_file):
            work_dir = os.path.dirname(from_file)
        else:
            work_dir = from_file
    else:
        work_dir = os.getcwd()
    work_dir = os.path.realpath(work_dir)
    os.chdir(work_dir)
    static_files = {
        '/static':  os.path.join(work_dir, 'static')
    }

    jam.context = Local()
    local_manager = LocalManager([jam.context])

    application = App(work_dir, load_task)
    if not testing:
        application = SharedDataMiddleware(application, static_files)
        application = local_manager.make_middleware(application)
    return application

class JamLogger(object):
    def __init__(self, app):
        self.app = app

    @cached_property
    def log(self):
        logger = logging.getLogger('jam_app')
        logger.setLevel(logging.DEBUG)
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        logger.addHandler(ch)

        return logger

    def debug(self, msg, *args, **kwargs):
        self.log.debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self.log.info(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.log.warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.log.error(msg, *args, **kwargs)

    def exception(self, msg, *args, **kwargs):
        self.log.exception(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        self.log.critical(msg, *args, **kwargs)

class App(object):
    def __init__(self, work_dir, load_task):
        mimetypes.add_type('text/cache-manifest', '.appcache')
        self.started = datetime.datetime.now()
        self.work_dir = work_dir
        self.state = consts.PROJECT_NONE
        self.__task = None
        self.privileges = None
        self._busy = 0
        self.pid = os.getpid()
        self.jam_dir = os.path.realpath(os.path.dirname(jam.__file__))
        self.jam_version = jam.version()
        self.__is_locked = 0
        self.application_files = {
            '/': self.work_dir,
            '/jam/': self.jam_dir
        }
        self.fileserver = SharedDataMiddleware(None, self.application_files, cache_timeout=1)
        self.url_map = Map([
            Rule('/', endpoint='root_file'),
            Rule('/<file_name>.html', endpoint='root_file'),
            Rule('/js/<file_name>', endpoint='file'),
            Rule('/css/<file_name>', endpoint='file'),
            Rule('/jam/js/<file_name>', endpoint='file'),
            Rule('/jam/js/ace/<file_name>', endpoint='file'),
            Rule('/jam/css/<file_name>', endpoint='file'),
            Rule('/jam/css/themes/<file_name>', endpoint='file'),
            Rule('/jam/img/<file_name>', endpoint='file'),
            Rule('/api', endpoint='api'),
            Rule('/upload', endpoint='upload')
        ])
        consts.app = self
        self.log = JamLogger(self)
        create_admin(self)
        self.build_id_prefix = '$buildID'
        self.save_build_id()
        if load_task:
            with self.admin.lock('$creating_task'):
                self.__task = create_task(self)
        self.check_migration()

    def create_task(self):
        result = None
        if self.state != consts.PROJECT_LOADING:
            self.state = consts.PROJECT_LOADING
            try:
                result = create_task(self)
                update_events_code(self.admin)
                consts.CLIENT_MODIFIED = False
                consts.SERVER_MODIFIED = False
                consts.write_settings()
                self.state = consts.RESPONSE
                result.__task_locked = True
            except ProjectNotCompleted:
                self.state = consts.PROJECT_NO_PROJECT
            except:
                self.state = consts.PROJECT_ERROR
                traceback.print_exc()
        return result

    @property
    def task(self):
        if not self.__task:
            self.__task = self.create_task()
        return self.__task

    def task_locked(self):
        return self.__is_locked > 0

    @property
    def __task_locked(self):
        pass

    @__task_locked.setter
    def __task_locked(self, value):
        if value:
            self.__is_locked += 1
        else:
            self.__is_locked -= 1

    def __call__(self, environ, start_response):
        jam.context.environ = environ
        jam.context.session = None
        request = JamRequest(environ)
        if consts.MAX_CONTENT_LENGTH > 0:
            request.max_content_length = 1024 * 1024 * consts.MAX_CONTENT_LENGTH
        adapter = self.url_map.bind_to_environ(request.environ)
        try:
            endpoint, values = adapter.match()
            if endpoint in ['file', 'root_file']:
                result = self.serve_file(environ, start_response, endpoint, **values)
                return result
            elif endpoint in ['api', 'upload']:
                response = getattr(self, 'on_' + endpoint)(request, **values)
        except HTTPException as e:
            if peek_path_info(environ) == 'ext':
                response = self.on_ext(request)
            else:
                response = e
        return response(environ, start_response)

    def check_modified(self, file_path, environ):
        if environ.get('HTTP_IF_MODIFIED_SINCE'):
            date1 = parse_date(environ['HTTP_IF_MODIFIED_SINCE'])
            date2 = datetime.datetime.utcfromtimestamp(os.path.getmtime(file_path)).replace(microsecond=0)
            if date1 != date2:
                try:
                    os.utime(file_path, None)
                except:
                    pass

    def serve_file(self, environ, start_response, endpoint, file_name=None):
        if endpoint == 'root_file':
            if file_name:
                file_name += '.html'
            if not file_name:
                file_name = 'index.html'
                environ['PATH_INFO'] = '/index.html'
            elif file_name == 'admin.html':
                file_name = 'builder.html'
            if file_name == 'index.html':
                self.check_modified(file_name, environ)
                self.check_project_modified()
            elif file_name == 'builder.html':
                if os.path.exists(file_name):
                    self.check_modified(file_name, environ)
                else:
                    self.check_modified(os.path.join(to_unicode(self.jam_dir, 'utf-8'), file_name), environ)
                    environ['PATH_INFO'] = '/jam/builder.html'
        if file_name:
            base, ext = os.path.splitext(file_name)
        init_path_info = None
        if consts.COMPRESSED_JS and ext and ext in ['.js', '.css']:
            init_path_info = environ['PATH_INFO']
            min_file_name = base + '.min' + ext
            environ['PATH_INFO'] = environ['PATH_INFO'].replace(file_name, min_file_name)
        try:
            try:
                return self.fileserver(environ, start_response)
            except Exception as e:
                if init_path_info:
                    environ['PATH_INFO'] = init_path_info
                    return self.fileserver(environ, start_response)
                else:
                    raise
        except Exception as e:
            return Response('')(environ, start_response)

    def create_post_response(self, request, result):
        response = Response()
        accepts_gzip = 0
        try:
            if request.environ.get("HTTP_ACCEPT_ENCODING").find("gzip") != -1:
                accepts_gzip = 1
        except:
            pass
        try:
            buff = json.dumps(result, default=json_defaul_handler)
        except:
            self.log.exception('wsgi.py create_post_response error')
            self.log.debug(result)
            raise
        response.headers['Content-Type'] = 'application/json'
        if accepts_gzip:
            buff = compressBuf(buff)
            response.headers['Content-encoding'] = 'gzip'
            response.headers['Content-Length'] = str(len(buff))
        response.set_data(buff)
        return response

    def get_client_address(self, request):
        try:
            return request.environ['HTTP_X_FORWARDED_FOR'].split(',')[-1].strip()
        except KeyError:
            return request.environ['REMOTE_ADDR']

    def create_session(self, request, task, user_info=None, session_uuid=None):
        if not user_info:
            user_info = {
                'user_id': None,
                'role_id': None,
                'role_name': '',
                'user_name': '',
                'admin': False
            }
        cookie = request.client_cookie
        session = {}
        session['ip'] = self.get_client_address(request)
        session['uuid'] = session_uuid
        session['user_info'] = user_info
        cookie['info'] = session
        cookie.modified = True
        return cookie

    def connect(self, request, task):
        if self.check_session(request, task):
            return consts.PROJECT_LOGGED

    def valid_session(self, task, session, request):
        if consts.SAFE_MODE:
            user_info = session['user_info']
            if not (user_info and user_info.get('user_id')):
                return False
            if not self.admin.ignore_change_ip and task != self.admin:
                ip = self.get_client_address(request)
                if not user_valid_ip(self.admin, user_info['user_id'], ip):
                    return False
            if not self.admin.ignore_change_uuid and task != self.admin:
                if not user_valid_uuid(self.admin, user_info['user_id'], session['uuid']):
                    return False
        return True

    def check_session(self, request, task):
        c = request.client_cookie
        if not c.get('info') and not consts.SAFE_MODE:
            c = self.create_session(request, task)
        session = c.get('info')
        if session:
            if not self.valid_session(task, session, request):
                self.logout(request, task)
                return False
            jam.context.session = session
            return True

    def default_login(self, task, login, password, ip, session_uuid):
        return login_user(self.admin, login, password, self.admin == task, ip, session_uuid)

    def login(self, request, task, form_data):
        time.sleep(0.3)
        ip = None
        session_uuid = None
        ip = self.get_client_address(request)
        session_uuid = str(uuid.uuid4())
        if self.admin == task or task.on_login is None:
            user_info = self.default_login(task, form_data['login'], form_data['password'], ip, session_uuid)
        elif task.on_login:
            try:
                try:
                    user_info = task.on_login(task, form_data, {'ip': ip, 'session_uuid': session_uuid})
                except:
                    # for compatibility with previous versions
                    if get_function_code(task.on_login).co_argcount == 5:
                        user_info = task.on_login(task, form_data['login'], form_data['password'], ip, session_uuid)
                    else:
                        raise

            except:
                user_info = None
                traceback.print_exc()
        if user_info:
            self.create_session(request, task, user_info, session_uuid)
            return True

    def logout(self, request, task):
        del request.client_cookie['info']
        jam.context.session = None

    def create_connection_pool(self):
        if self.task:
            self.__task_locked = False
            try:
                self.task.create_pool()
            finally:
                self.__task_locked = True

    def get_privileges(self, role_id):
        if self.privileges is None:
            roles, privileges = get_roles(self.admin)
            self.privileges = privileges
        try:
            result = self.privileges[role_id]
        except:
            result = {}
        return result

    def init_client(self, task):
        session = jam.context.session
        priv = None
        user_info = {}
        if session:
            user_info = session['user_info']
            role_id = user_info.get('role_id')
            if role_id:
                priv = self.get_privileges(role_id)
        result = {
            'task': task.get_info(),
            'settings': consts.settings,
            'locale': consts.locale,
            'language': consts.lang,
            'user_info': user_info,
            'privileges': priv
        }
        return result, ''

    def check_migration(self):
        path = os.path.join(self.work_dir, 'migration')
        files = []
        if os.path.exists(path):
            for file_name in os.listdir(path):
                files.append(os.path.join(self.work_dir, 'migration', file_name))
        files_len = len(files)
        if files_len:
            if files_len == 1:
                self.import_md(files[0], False)
            else:
                self.log.error('More than one file in migration folder')

    def import_metadata(self, task, task_id, file_name, from_client): #for compatibility with previous versions
        self.import_md(file_name, from_client)

    def import_md(self, file_name, from_client):
        if not self.under_maintenance:
            with self.admin.lock('$metadata_import'):
                consts.MAINTENANCE = True
                consts.PARAMS_VERSION += 1
                consts.write_settings()
                self.save_build_id()
                self.__task_locked = False
                try:
                    result = import_metadata(self.admin, file_name, from_client)
                    success, error, message = result
                    if success and self.task:
                        reload_task(self)
                        update_events_code(self.admin)
                        self.privileges = None
                except:
                    self.log.exception('wsgi.py create_post_response error')
                finally:
                    self.__task_locked = True
                    consts.MAINTENANCE = False
                    consts.PARAMS_VERSION += 1
                    if success:
                        consts.BUILD_VERSION += 1
                        consts.MODIFICATION += 1
                    consts.write_settings()
                    self.save_build_id()
                return result

    def get_under_maintenance(self):
        return consts.MAINTENANCE

    under_maintenance = property(get_under_maintenance)

    def __get_client_modified(self):
        return consts.CLIENT_MODIFIED

    def __set_client_modified(self, value):
        consts.CLIENT_MODIFIED = value
        consts.MODIFICATION += 1
        consts.PARAMS_VERSION += 1
        consts.write_settings()
        self.save_build_id()

    client_modified = property(__get_client_modified, __set_client_modified)

    def __get_server_modified(self):
        return consts.SERVER_MODIFIED

    def __set_server_modified(self, value):
        consts.SERVER_MODIFIED = value
        consts.MODIFICATION += 1
        consts.PARAMS_VERSION += 1
        consts.write_settings()
        self.save_build_id()

    server_modified = property(__get_server_modified, __set_server_modified)

    def check_project_modified(self):
        if self.task:
            with self.admin.lock('$code_updating'):
                params = consts.read_params(['CLIENT_MODIFIED', 'SERVER_MODIFIED', 'MAINTENANCE'])
                if not params['MAINTENANCE'] and (params['CLIENT_MODIFIED'] or params['SERVER_MODIFIED']):
                    self.__task_locked = False
                    try:
                        if params['SERVER_MODIFIED']:
                            reload_task(self)
                            consts.BUILD_VERSION += 1
                        if params['CLIENT_MODIFIED']:
                            update_events_code(self.admin)
                        consts.CLIENT_MODIFIED = False
                        consts.SERVER_MODIFIED = False
                        consts.MODIFICATION += 1
                        consts.PARAMS_VERSION += 1
                        consts.write_settings()
                        self.save_build_id()
                    finally:
                        self.__task_locked = True

    @property
    def build_id(self):
        return '%s_%s_%s' % (self.build_id_prefix, consts.BUILD_VERSION, consts.PARAMS_VERSION)

    def save_build_id(self):
        with self.admin.lock('$save_build_id'):
            path = os.path.join(self.work_dir, 'locks')
            for file_name in os.listdir(path):
                if file_name.startswith(self.build_id_prefix):
                    os.remove(os.path.join(path, file_name))
            file_write(os.path.join(path, self.build_id), '')

    def check_build(self):
        path = os.path.join(self.work_dir, 'locks')
        if not os.path.exists(os.path.join(path, self.build_id)):
            with self.admin.lock('$build_checking'):
                cur_build_version = consts.BUILD_VERSION
                cur_params_version = consts.PARAMS_VERSION
                build_version = None
                params_version = None
                for file_name in os.listdir(path):
                    if file_name.startswith(self.build_id_prefix):
                        cur_build_id = file_name
                        arr = cur_build_id.split('_')
                        build_version = int(arr[1])
                        params_version = int(arr[2])
                        break
                if params_version != cur_params_version:
                    consts.read_settings()
                if build_version != cur_build_version:
                    self.__task_locked = False
                    try:
                        reload_task(self)
                    finally:
                        self.__task_locked = True
                    consts.read_settings()

    def on_api(self, request):
        error = ''
        if request.method == 'POST':
            r = {'result': None, 'error': None}
            try:
                data = request.get_data()
                if type(data) != str:
                    data = to_unicode(data, 'utf-8')
                method, task_id, item_id, params, modification = json.loads(data)
                if task_id == 0:
                    task = self.admin
                else:
                    task = self.task
                    if not task:
                        task = self.task
                        if not task:
                            lang = consts.lang
                            result = {'status': None, 'data': {'error': lang['error'], \
                                'info': lang['info']}, 'modification': None}
                            result['status'] = self.state
                            if self.state == consts.PROJECT_LOADING:
                                result['data']['project_loading'] = lang['project_loading']
                            elif self.state == consts.PROJECT_NO_PROJECT:
                                result['data']['no_project'] = lang['no_project']
                            elif self.state == consts.PROJECT_ERROR:
                                result['data']['project_error'] = lang['project_error']
                            r ['result'] = result
                            return self.create_post_response(request, r)
                if not task:
                    result = {'status': consts.PROJECT_NO_PROJECT, 'data': None, 'modification': None}
                else:
                    request.task = task
                    self.check_build()
                    result = {'status': consts.RESPONSE, 'data': None, 'modification': consts.MODIFICATION}
                    if task_id and modification and modification != consts.MODIFICATION:
                        result['status'] = consts.PROJECT_MODIFIED
                    elif self.under_maintenance:
                        result['status'] = consts.PROJECT_MAINTAINANCE
                    elif method == 'connect':
                        self.connect(request, task)
                        result['data'] = self.connect(request, task)
                    elif method == 'login':
                        result['data'] = self.login(request, task, params[0])
                    elif method == 'logout':
                        self.logout(request, task)
                        result['status'] = consts.PROJECT_NOT_LOGGED
                        result['data'] = consts.PROJECT_NOT_LOGGED
                    else:
                        if not self.check_session(request, task):
                            result['status'] = consts.PROJECT_NOT_LOGGED
                            result['data'] = consts.PROJECT_NOT_LOGGED
                        else:
                            item = task
                            if task and item_id:
                                item = task.item_by_ID(item_id)
                            self._busy += 1
                            try:
                                data = self.get_response(item, method, params)
                            finally:
                                self._busy -= 1
                            result['data'] = data
                r ['result'] = result
            except AbortException as e:
                traceback.print_exc()
                error = error_message(e)
                r['result'] = {'data': [None, error]}
                r['error'] = error
            except Exception as e:
                traceback.print_exc()
                error = error_message(e)
                if consts.DEBUGGING and task_id != 0:
                    raise
                r['result'] = {'data': [None, error]}
                r['error'] = error
            response = self.create_post_response(request, r)
            request.save_client_cookie(response, self, task)
            return response

    def get_response(self, item, method, params):
        if method == 'open':
            return item.select_records(params, safe=True)
        elif method == 'apply':
            return item.apply_changes(params, safe=True)
        elif method == 'server':
            return self.server_func(item, params[0], params[1])
        elif method == 'print':
            return item.print_report(*params, safe=True), ''
        elif method == 'load':
            return self.init_client(item)

    def server_func(self, obj, func_name, params):
        result = None
        error = ''
        func = getattr(obj, func_name)
        if func:
            result = func(obj, *params)
        else:
            raise Exception('item: %s no server function with name %s' % (obj.item_name, func_name))
        return result, error

    def on_ext(self, request):
        if request.method == 'POST':
            r = {'result': None, 'error': None}
            method = get_path_info(request.environ)
            data = request.get_data()
            if type(data) != str:
                data = to_unicode(data, 'utf-8')
            try:
                params = json.loads(data)
            except:
                params = None
            if self.task:
                try:
                    data = None
                    if self.under_maintenance:
                        status = consts.UNDER_MAINTAINANCE
                    elif self.task.on_ext_request:
                        status = consts.RESPONSE
                        self._busy += 1
                        try:
                            data = self.task.on_ext_request(self.task, method, params)
                        finally:
                            self._busy -= 1
                    else:
                        status = None
                    r['result'] = {'status': status, 'data': data, 'modification': consts.MODIFICATION}
                except AbortException as e:
                    traceback.print_exc()
                    r['result'] = {'data': [None, error_message(e)]}
                    r['error'] = error_message(e)
                except Exception as e:
                    traceback.print_exc()
                    r['result'] = {'data': [None, error_message(e)]}
                    r['error'] = error_message(e)
            else:
                r['result'] = {'status': self.state, 'data': None, 'modification': None}
            return self.create_post_response(request, r)

    def on_upload(self, request):
        if request.method == 'POST':
            r = {'result': None, 'error': None}
            task_id = int(request.form.get('task_id'))
            item_id = int(request.form.get('item_id'))
            field_id = int(request.form.get('field_id'))
            path = request.form.get('path')
            if task_id == 0:
                task = self.admin
            else:
                task = self.task
            if task:
                request.task = task
                result = {'status': consts.RESPONSE, 'data': None, 'modification': consts.MODIFICATION}
                r ['result'] = result
                if not self.check_session(request, task):
                    r['result']['status'] = consts.NOT_LOGGED
                    r['result']['data'] = consts.NOT_LOGGED
                else:
                    f = request.files.get('file')
                    file_name = request.form.get('file_name')
                    if f and file_name:
                        base, ext = os.path.splitext(file_name)
                        ext = ext.lower()
                        upload_result = None
                        if task.on_upload:
                             upload_result = task.on_upload(task, path, file_name, f)
                        if upload_result:
                            path, file_name = upload_result
                            r['result']['data'] = {'file_name': file_name, 'path': path}
                        else:
                            if item_id != -1 and field_id != -1:
                                item = task.item_by_ID(item_id)
                                field = item.field_by_ID(field_id)
                                if field.data_type == consts.IMAGE:
                                    if not valid_uploaded_file('image/*', ext):
                                        r['error'] = consts.lang['upload_not_allowed']
                                elif field.data_type == consts.FILE:
                                    if not valid_uploaded_file(field.field_file['accept'], ext):
                                        r['error'] = consts.lang['upload_not_allowed']
                                else:
                                    r['error'] = 'Operation prohibited'
                            else:
                                if not ext in consts.upload_file_ext:
                                    r['error'] = '%s - %s' % (request.form.get('file_name'), consts.lang['upload_not_allowed'])
                                    self.log.error(r['error'])
                            file_name = ('%s_%s%s') % (base, datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S_%f'), ext)
                            file_name = secure_filename(file_name)
                            file_name = file_name.replace('?', '')
                            if task_id == 0:
                                path = os.path.join('static', 'builder')
                            else:
                                path = os.path.join('static', 'files')
                            if not r['error']:
                                dir_path = os.path.join(to_unicode(self.work_dir, 'utf-8'), path)
                                if not os.path.exists(dir_path):
                                    os.makedirs(dir_path)
                                f.save(os.path.join(dir_path, file_name))
                                r['result']['data'] = {'file_name': file_name, 'path': path}
                    else:
                        r['error'] = 'File upload invalid parameters'
            else:
                r['result'] = {'status': self.state, 'data': None, 'modification': None}
            return self.create_post_response(request, r)

    def stop(self, sigvalue):
        self.kill()

    def kill(self):
        import signal, subprocess
        if os.name == "nt":
            subprocess.Popen("taskkill /F /T /pid %i" % self.pid, shell=True)
        else :
            os.killpg(self.pid, signal.SIGKILL)
